#include "../include/hash_table.h"
#include "../include/except.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

const Exception out_of_memory = { "Out Of Memory" }; // OOM Exception.

struct Entry{
    const char *key;
    void *obj;
    struct Entry *next;
};
struct Hash_Table{
    uint32_t size;
    hash_func *hash;
    cleanup_func *cf;
    Entry **entries;
};
static size_t hash_table_index(Hash_Table *ht, const char *key){
    size_t res = (ht->hash(key,strlen(key)) % ht->size);
    return res;

}

Hash_Table *hash_table_new(uint32_t size, hash_func *hf, cleanup_func *cf, int flags){
    // Allocate table and populate options.
    Hash_Table *ht = malloc(sizeof(*ht));    
    // Error Check
    if(!ht){
       if(!(flags & SOFTFAIL)) RAISE(out_of_memory);
       return NULL; 
    }
    // Populate options
    ht->size = size;
    ht->hash = hf;
    if(cf) ht->cf = cf;
    ht->cf = free; 
    // Zero Out Entries
    ht->entries = calloc(sizeof(Entry*), ht->size);
    // Error check
    if(!ht->entries){
       if(!(flags & SOFTFAIL)) RAISE(out_of_memory);
       return NULL; 
    }
    // Allocation succeded return table.
    return ht;
}
void hash_table_free(Hash_Table *ht){
    for(uint32_t i = 0; i < ht->size; i++){
        while(ht->entries[i]){
            Entry *tmp = ht->entries[i];
            ht->entries[i] = ht->entries[i]->next;
            ht->cf(tmp->obj);
            free((void*)tmp->key); 
            free(tmp);
        }
    }
    free(ht->entries); 
    free(ht);
}
void hash_table_print(Hash_Table *ht, int printNull){
    printf("--- Hash Table ---\n");
    for (uint32_t i = 0; i < ht->size; i++){
        if(ht->entries[i] == NULL){
            // Null Entry
            if(printNull) printf("\t--- %i ---\t = EMPTY\n",i);
        } else {
            Entry *tmp = ht->entries[i];
            while(tmp != NULL){
                printf("\t--- %i ---\t %s \t(%p)\n", i, tmp->key, tmp->obj);
                tmp = tmp->next;
            }
        }
    }
    printf("--- End Table --- \n");
}
uint64_t hash_table_create(Hash_Table *ht, const char *key, void *obj,int flags){
    // Null Check 
    if(key == NULL || obj == NULL || ht == NULL) return 0;
    
    // Only creation, no updating must delete.
    size_t i = hash_table_index(ht,key);
    if(hash_table_read(ht, key) != NULL) return 0;
    
    // Allocate Entry.
    Entry *e = malloc(sizeof(*e));
    // Error check.
    if(!e) {
        if(!(flags & SOFTFAIL)) RAISE(out_of_memory);
        return 0;
    }
    // Copy key for ownership.  
    e->key = strdup(key);
    // Error check.
    if(!e) {
        if(!(flags & SOFTFAIL)) RAISE(out_of_memory);
        return 0;
    }
    // Populate Entry.
    e->obj = obj;
    e->next = ht->entries[i];
    ht->entries[i] = e;
    // Return the index of the entry. 
    return i;
}
void* hash_table_read(Hash_Table *ht, const char *key){
    // NULL Checks.
    if(key == NULL || ht == NULL) return NULL;
    // Get index of entry.
    size_t i = hash_table_index(ht,key);
    Entry *e = ht->entries[i];
    // Traverse list until you find the key.
    while(e != NULL && strcmp(e->key,key) != 0){
        e = e->next;
    }
    // Didn't find anything.
    if (e == NULL) return NULL;
    // Found something return object to caller. 
    return e->obj;
}
void* hash_table_delete(Hash_Table *ht, const char *key){
    // NULL checks.
    if(key == NULL || ht == NULL) return NULL;
    // Get index of entry for search.
    size_t i = hash_table_index(ht,key);
    Entry *e = ht->entries[i];
    Entry *prev = NULL;
    // Traverse list until you find the key.
    while(e != NULL && strcmp(e->key,key) != 0){
        prev = e;
        e = e->next;
    }
    // Nothing found
    if( e == NULL ) return NULL;
    // Found something, reconstruct list if in the middle. 
    if(prev == NULL){
        ht->entries[i] = e->next;
    } else {
        prev->next = e->next;
    }
    // Return the obj pointer to caller and free entry.
    void *res = e->obj; 
    free(e);
    return res;
}

uint64_t hash_fnv0(const char *buf, size_t len){
    uint64_t result = 0;
    for(size_t i = 0; i < len; i++){
        result *= FNV_PRIME;
        result ^= buf[i];
    }
    return result;
}
uint64_t hash_fnv1(const char *buf, size_t len){
    uint64_t result = FNV_OFFSET;
    for(size_t i = 0; i < len; i++){
        result *= FNV_PRIME;
        result ^= buf[i];
    }
    return result;
}
uint64_t hash_fnv1a(const char *buf, size_t len){
    uint64_t result = FNV_OFFSET;
    for(size_t i = 0; i < len; i++){
        result ^= buf[i];
        result *= FNV_PRIME;
    }
    return result;
}
